# import the data that we've analyzed and prepared for model building
merged.data <- readRDS(file = "data/merged_data_prepared.RData")
# let's remind ourselves about the data structure
str(merged.data)
summary(merged.data)

##########################################################
## ADDITIONAL DATA PREPARATION: IMPUTING MISSING VALUES ##
##########################################################

# Before applying the NB algorithm, let's first handle the missing values we've spotted
# previously (otherwise, those observations will be removed during the model building phase); 
# we've already handled the missing value of the Fare and Embarked variables, 
# now, we'll try to handle the Age variable

# first, identify the observations with missing Age value
age.missing <- which( is.na(merged.data$Age) )
length(age.missing)

# now, build a prediction model (namely, a regression tree)
# note that the method is "anova" as we are building now a regression (not classification) tree
# also note that we restricted the dataset to those instances that are not missing the Age value
# the minus sign (-) means 'all indices not present in the missing.indices vector'
require(rpart)
require(rpart.plot)
age.model <- rpart(Age ~ Pclass + Title + FamilySize, 
                   data = merged.data[ -age.missing, ], method = "anova")
# plot the resulting regression tree
prp(age.model, type = 3, extra = 1)
library(rattle)
library(RColorBrewer)

fancyRpartPlot(age.model)

# let's use the built model to predict the missing Age values and add them to our dataset
# first, predict Age value using the created model
missing.predicted <- predict(object = age.model, 
                             newdata = merged.data[age.missing,] )

# store the values of the Age variable before imputation in an auxiliary data frame
# so that after imputation we can draw plot (with histograms) that will help us assess 
# the effect of the imputation on the distribution of the Age variable
ageWithNAs <- data.frame(age=merged.data$Age)
ageWithNAs$lbl <- 'before'
#ggplot(ageWithNAs, aes(x=age)) + geom_density()

# add predicted values to the dataset
merged.data$Age[ age.missing ] <- missing.predicted 
# check that there are no more missing Age values
summary( merged.data$Age )
summary(ageWithNAs$age)

# check how the imputation affected the distribution of the Age attribute
ageNoNAs <- data.frame(age=merged.data$Age)
ageNoNAs$lbl <- 'after'
ageData <- rbind(ageWithNAs, ageNoNAs)
require(ggplot2)
ggplot(ageData, aes(x=age, fill=lbl)) + geom_density(alpha=0.2) + theme_bw()
# it seems that the missing age values were primarily of those passengers of 27-30 years of age
# let's check that by inspecting the statistics of the predicted age values
summary(missing.predicted)

#################################################
## ADDITIONAL DATA PREPARATION: DISCRETIZATION ##
#################################################

# NB is often used with nominal variables; if numeric variables are to be used, they have to
# be either 1) represented as probabilities based on the appropriate probability distribution 
# (typically Normal dist.), or 2) discretized 
# In this case, we have two numeric variables - Age and Fare - and have to handle them in one
# of these two ways

# let's check if the two numeric attributes are normally distributed
shapiro.test( merged.data$Age )
# p-value = 9.468e-12 indicates that the attribute is far from being normally dist.
shapiro.test( merged.data$Fare )
# the same applies to the Fare attribute (p-value < 2.2e-16)

# so, we'll follow the option of discretizing the two numeric attributes
# select the two numeric attributes that are to be discretized; include also 
# the class variable since it is required for supervised discretization
merged.numeric <- merged.data[, c("Age", "Fare", "Survived")]

#############################
# unsupervised discretization
#############################
# install.packages("bnlearn")
library(bnlearn)

## unsupervised discretization divides a continuous feature into groups (bins) 
## without taking into account the class attribute or any other information. 
## it can be done in 2 ways: equal length intervals and equal frequency intervals
## the 2nd option tends to be better of the two
d <- discretize(data = merged.numeric[,c(1,2)], method = "quantile", breaks = 5, ordered = T)
summary(d$Fare)
merged.data$AgeUSDiscr <- d$Age
merged.data$FareUSDiscr <- d$Fare

###########################
# supervised discretization
###########################
# install.packages('smbinning')
library(smbinning)

# supervised discretization method is based on 2 key ideas:
# - values should be split into intervals that maximize the difference between the 
#   groups defined by the class attribute (in this case Survived/Not survived)
# - partitioning should not be too fine-grained, to avoid over-fitting

# smbinning methods requires that the class variable is binary numeric variable with values 0/1 
merged.numeric$Survived <- as.numeric(merged.numeric$Survived)
merged.numeric$Survived <- merged.numeric$Survived - 1
# discretize Age attribute
ageDiscr <- smbinning(df = merged.numeric, y = "Survived", x = "Age") 
# update the merged.numeric df with the discretized Age
merged.numeric <- smbinning.gen(merged.numeric, ageDiscr, "AgeDiscr")
# discretize Fare attribute
fareDiscr <- smbinning(df = merged.numeric, y = "Survived", x = "Fare")
# update the merged.numeric df with the discretized Fare
merged.numeric <- smbinning.gen(df = merged.numeric, fareDiscr, "FareDiscr")
str(merged.numeric)

# add these discretized variables to our dataset
merged.data$AgeSDiscr <- merged.numeric$AgeDiscr
merged.data$FareSDiscr <- merged.numeric$FareDiscr



#####################################################################
## SPLITTING DATA INTO TRAIN AND VALIDATION SETS (ALREADY COVERED) ##
#####################################################################
require(caret)
set.seed(1010)
# stratified random sampling 
# select 80% of all the observations for training and the rest for validation
train.data <- merged.data[1:891,]
train.index <- createDataPartition(train.data$Survived, p = .80,
                                   list = FALSE)
train <- train.data[train.index,]
validate <- train.data[-train.index,]

################################
######  MODEL BUILDING    ######
################################
require(e1071)

nb.model <- naiveBayes(Survived ~ Pclass + Sex + Embarked + FamilySize + Title + 
                             # AgeUSDiscr + FareUSDiscr,
                              AgeSDiscr + FareSDiscr,
                       data = train)

# now, let's do the prediction using the created model 
# note: eps parameter defines an epsilon-range to apply laplace smoothing, that is, to replace
# zero or close-zero probabilities by the theshold
nb.pred <- predict(object = nb.model, newdata = validate, 
                   type = "class", eps = 0, threshold = 0.001)
nb.pred[1:10]

# f. for computing several evaluation metrics:
# accuracy, precision, recall, F-measure, sensitivity, specificity
compute.eval.measures <- function(cm) {
  n <- sum(cm)
  acc <- sum(diag(cm))/n
  
  # since we're interested in predicting survival, we'll compute precision and recall
  # using values of the confusion matrix that are related to the Yes outcome
  TP <- cm[2,2]
  TN <- cm[1,1]
  FP <- cm[1,2]
  FN <- cm[2,1]
  prec <- TP / (TP + FP) 
  rec <- TP / (TP + FN)
  f1 <- (2*prec*rec)/(prec+rec)
  TNR <- TN/(TN+FP)
  # return the performance metrics
  c(accuracy = acc, precision = prec, recall = rec, F1 = f1, sensitivity = rec, specificity = TNR)  
}

validate.cm <- table(actual=validate$Survived, predicted=nb.pred)
validate.cm
validate.metrics <- compute.eval.measures(validate.cm)
validate.metrics

# not that bad 
# similar (even worse) results (accuracy=0.75122) have been reported here: 
# http://sqldatamine.blogspot.com/2013/07/catastrophic-test-of-naive-bayes.html
# so, we should consider one (or more) of the following options:
# - exclude some of the used features
# - introduce new features
# - use another discretization method
# - choose another algorithm (NB might not be the best solution for this particular problem)


##########################
######  ROC CURVES  ######
##########################

# when doing class prediction based on a built prediction model, 
# we do not need to apply the "majority class" principle to do the classification 
# (that is, use the 0.5 probability threshold to choose the class); 
# we can also do the classification based on the chosen threshold value
# e.g., we can set the probability threshold to 0.8 and say that we will consider 
# that a passanger survived if the predicted probability of survival is >=0.8

# the question is how to choose proper threshold value (i.e., the cutoff point)
# ROC curves can help with that
## <! explain ROC curves, AUC, sensitivity and specificity !>
# Sensitivity (the true positive rate) measures the proportion of positives that are 
# correctly identified as such; the same as recall; it is calculate as TP/(TP+FN)
# Specificity (the true negative rate) measures the proportion of negatives that are 
# correctly identified as such; it is calculate as TN/(TN+FP)

# to create ROC curves, we need to install and load the pROC package 
# http://web.expasy.org/pROC/
# install.packages("pROC")
library(pROC)

# let's start by making predictions in such a way that we predict just the probabilities
# of the two possible values of the outcome (the Survived variable)  
nb.pred.prob <- predict(object = nb.model, newdata = validate, type = "raw", 
                        eps = 0.001, threshold = 1)
# note that the type parameter is set to 'raw'
# let's inspect the first ten predictions, to see what they look like
nb.pred.prob[1:10,]

# now, compute the parameters for the ROC curve 
nb.roc <- roc(response = validate$Survived, predictor = nb.pred.prob[,2] )
# note that nb.pred.prob[,2] means that we are taking the probabilities given in the 2nd col.
# of the matrix, that is, the probabilities of the 'Yes' value of the Survived variable
# since it is the 'positive class' (the class we are interested in predicting) 
# in the context of this problem

# we can check the Area Under the Curve (AUC) value 
nb.roc$auc
# it's rather good, as it is close to 1 (0.8567)

# we can also plot the ROC curve
plot.roc(validate$Survived, 
         nb.pred.prob[,2], 
         ci=T,  # compute conf. inter. (of AUC by default)  
         print.auc=T) # also print AUC and its conf. inter. on the plot 

# let's now use the ROC curve to determine the threshold; to that end, we'll use the coords f.
?coords
# see the documentation for the meaning of the parameters 'x' and 'best.method'
# with the 'ret' (return) parameter we chose the metrics to return - 
# other metrics can be returned, as well (see the documentation)
# prevalence and FN cost are two parametrs that we might want to specify:
# - prevalence is particularly important if observations are not uniformy distributed across classes
# - FN cost is important if we want to favour FN over FP or vice versa
# prevalence is the proportion of instances of the positive class
prevalence <- as.numeric(table(validate$Survived)[2]/nrow(validate))
# we reduce the cost of having FN if we want to favor the reduction in FP
fn.cost <- 1 
nb.coords <- coords(roc = nb.roc, x = "best", best.method = "closest.topleft", 
                    best.weights = c(fn.cost, prevalence),
                    ret = c("threshold", "accuracy", "sensitivity", "specificity"))
print(nb.coords)
threshold <- as.numeric( nb.coords[1] )

# let's now use the computed threshold to assign class labels to instances
nb.pred.labels <- sapply(nb.pred.prob[,2], 
                              function(x) { ifelse(test = (x >= threshold), yes = "Yes", no = "No")})
validate.cm2 <- as.matrix( table( true = validate$Survived, predicted = nb.pred.labels ) )
validate.cm2
compute.eval.measures( validate.cm2 )


# We can also use ROC curves to (visually) compare different classifiers
# To illustrate that, let's compare the DT and NB classifiers
# we'll build DT classifier with the same set of features
dt.model <- rpart(Survived ~ Pclass + Sex + Embarked + FamilySize + Title +
                    AgeUSDiscr + FareUSDiscr, 
                    data = train, method = "class")
dt.pred.prob <- predict(object = dt.model, newdata = validate, type = "prob")
head(dt.pred.prob)

# plot the ROC curves
roc.dt <- plot.roc(validate$Survived, dt.pred.prob[,2],  
                   main="Comparison of DT and NB classifiers", 
                   col="#1c61b6")  
roc.nb <- lines.roc(validate$Survived, nb.pred.prob[,2], col="#008600")  
legend("bottomright", legend=c("DT", "NB"), col=c("#1c61b6", "#008600"), lwd=2)

# we can do a test to determine if there is a statisticaly significant difference 
# between the performance of the two classifiers
# the comparison is based on the AUC values of the two classifiers
# the null hypothesis of the test is that the AUC values of the two classifiers 
# do not significantly differ 
test.res <- roc.test(roc.dt, roc.nb)  
c(Z=as.numeric(test.res$statistic), p=as.numeric(test.res$p.value))
# the two classifiers are not significanty differnet in their performance (measured by AUC)